//
//   Defines a set of macros that simulate hanging cables where the 'gravity'
//   points away from the origin.
//

#declare g_debug_cable = 0;

#declare g_cable_rand = seed(0);

#macro macro_cable_rand(p_max)
    (rand(g_cable_rand) * 0.99999999 * p_max)
#end

#macro macro_cable_rand_length(p_max, p_min)
    // Pick a random number from -14 to 2, raise e to the power of it, and
    // multiply the length by it.  Add a minimum value +/- 5%
    #local l_rand = rand(g_cable_rand);
    ((exp((l_rand * 18) - 16) * p_max) + p_min)
#end

#macro macro_snapped_cable(p_point, p_length, p_radius)
    #if (p_length > 0)
        #local l_unit = (p_point * <0, 1, 1>) / vlength(p_point);
        #local l_to = p_point + l_unit * p_length;
        cylinder {
            p_point, l_to, p_radius
        }
    #end
#end

#macro macro_sagging_cable(p_left, p_right, p_sag, p_radius)
    #if (vlength(p_left - p_right) > 0.000001)
        //  The sagging cable consists of a spline that follows the curve of an
        //  approximated hyperbolic cosine. 
        //
        //  Gravity is the unit vector of the direction from the origin to the
        //  point in question.

        //
        // To get different amounts of realistic cable sagging, we need to
        // create a virtual x-y axis and figure out the approriate starting and
        // finishing X values for a given sag amount.  This is made more
        // difficult given that the Y offset for the X values can be different,
        // eg, cable hanging off a sloping rail.  Then once we've figured out
        // our starting X values, we can figure out the Y offset for each
        // segment of the curve.  We then map the virtual X offset back to a
        // real world vector representing that offset, scaling the Y
        // appropriately, and then add to that vector the the Y offset
        // multiplied by the unit vector.  There must be an easier way of doing
        // this, but it's probably not as *fun*.  I derived the following very
        // late at night and will be quite surprised if it actually works...
        //


        // Adapt the number of segments to the distance from the observer
        // to save on the number of objects used
        #ifdef (g_observer_position)
            // The documentation says I can't trust pow(x) for non integers,
            // so I'll use logarithms to get x^(1/4)
            #local l_cable_segments = 40 - 12 * exp(1/4*ln(2 * vlength(g_observer_position - p_left)));
        #else
            #local l_cable_segments = 40;
        #end

        #local l_va = p_left;
        #local l_vb = p_right;
        #local l_vc = l_vb - l_va;
        #local l_vc_inc = l_vc / l_cable_segments;
        #local l_vgrav = (l_va + l_vb) / 2;
        #local l_vgrav = l_vgrav * <0, 1, 1>;
        #local l_vgrav = l_vgrav / vlength(l_vgrav);
        #local l_theta = VAngle(l_vgrav, l_vc);
        #local l_u = vlength(l_vc) * sin(l_theta);
        #local l_v_signed = vlength(l_vc) * cos(l_theta);
        #local l_v = abs(l_v_signed);
        #local l_v_inc = l_v / l_cable_segments;
        #local l_s = p_sag - l_v / 2;


        #if (l_s < 0)
            #local l_constant = l_v / 2 - p_sag + 1;
            #local l_u1 = ln(l_constant +
                    sqrt(f_sqr(l_constant) - 1));
            #local l_u2 = ln((l_constant + l_v) +
                    sqrt(f_sqr(l_constant + l_v) - 1));

            #if (l_v_signed > 0)
                #local l_temp = l_u1;
                #local l_u1 = l_u2;
                #local l_u2 = l_temp;
            #end
        #else
            #local l_constant = l_s + 1;
            #local l_u1 = -ln(l_constant +
                    sqrt(f_sqr(l_constant) - 1));
            #local l_u2 = ln((l_constant + l_v) +
                    sqrt(f_sqr(l_constant + l_v) - 1));
    
            #if (l_v_signed > 0)
                #local l_temp = l_u1;
                #local l_u1 = l_u2;
                #local l_u2 = l_temp;
            #end
        #end

        #local l_ux_inc = (l_u2 - l_u1) / l_cable_segments;
        #local l_vx_inc = (cosh(l_u2) - cosh(l_u1)) / l_cable_segments;

        #local l_i = 0;
        #while (l_i < l_cable_segments)
            #local l_ux = l_u1 + l_ux_inc * l_i;
            #local l_vx = cosh(l_u1) + l_vx_inc * l_i;
            #local l_distance = l_vx - cosh(l_ux);
            #local l_displacement = l_vgrav * l_distance;
            #local l_ve_i = l_va + l_vc_inc * l_i;
            #local l_begin = l_ve_i + l_displacement;

            #local l_ux = l_u1 + l_ux_inc * (l_i + 1);
            #local l_vx = cosh(l_u1) + l_vx_inc * (l_i + 1);
            #local l_distance = l_vx - cosh(l_ux);
            #local l_displacement = l_vgrav * l_distance;
            #local l_ve_i = l_va + l_vc_inc * (l_i + 1);
            #local l_end = l_ve_i + l_displacement;

            sphere {
                l_begin, p_radius
            }

            cylinder {
                l_begin, l_end, p_radius
            }

            #local l_i = l_i + 1;
        #end
    #end
#end

#macro macro_build_cable3(p_left, p_middle, p_right, p_sag1, p_radius)
    #local l_fail = macro_cable_rand(100);
    #local l_snap_pct = 1;
    #local l_mfail_pct = 3;
    #local l_gone_pct = 3;

    // Do snapped cables first
    #if (l_fail < l_snap_pct)

        // Figure out the cable lengths
        #local l_cable_length = vlength(p_left - p_right) / 2;
        #local l_snap_length_1 = macro_cable_rand(l_cable_length);
        #local l_snap_length_2 = l_cable_length - l_snap_length_1;

        #local l_random = macro_cable_rand(7);
        #if (l_random < 2.25)
            macro_snapped_cable(p_left, l_snap_length_1, p_radius)
        #end
        #if ((l_random > 1.75) & (l_random < 4))
            macro_snapped_cable(p_right, l_snap_length_2, p_radius)
        #end
        #if ((l_random > 4) & (l_random < 5))
            macro_snapped_cable(p_middle, l_snap_length_2, p_radius)
        #end
    #end
    #if ((l_fail > l_snap_pct) & (l_fail < (l_snap_pct + l_mfail_pct)))
        #local l_sag = macro_cable_rand_length(2 * p_sag1, 2 * p_sag1/5);
        macro_sagging_cable(p_left, p_right, l_sag, p_radius)
    #end
    #if (l_fail > (l_snap_pct + l_mfail_pct + l_gone_pct))
        #local l_sag = macro_cable_rand_length(p_sag1, p_sag1/5);
        macro_sagging_cable(p_left, p_middle, l_sag, p_radius)
        #local l_sag = macro_cable_rand_length(p_sag1, p_sag1/5);
        macro_sagging_cable(p_middle, p_right, l_sag, p_radius)
    #end
#end

#macro macro_build_cable2(p_left, p_right, p_sag1, p_radius)
    #local l_fail = macro_cable_rand(100);
    #local l_snap_pct = 2;
    #local l_gone_pct = 5;

    // Do snapped cables first
    #if (l_fail < l_snap_pct)

        // Figure out the cable lengths
        #local l_cable_length = vlength(p_left - p_right);
        #local l_snap_length_1 = macro_cable_rand(l_cable_length);
        #local l_snap_length_2 = l_cable_length - l_snap_length_1;

        #local l_random = macro_cable_rand(3);
        #if (l_random < 2)
            macro_snapped_cable(p_left, l_snap_length_1, p_radius)
        #end
        #if (l_random > 1)
            macro_snapped_cable(p_right, l_snap_length_2, p_radius)
        #end
    #end
    #if (l_fail > (l_snap_pct + l_gone_pct))
        #local l_sag = macro_cable_rand_length(p_sag1, p_sag1/5);
        macro_sagging_cable(p_left, p_right, l_sag, p_radius)
    #end
#end
